import logging
import os

import yaml

from connection import Connection
from bucket import BucketFactory

requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)

log = logging.getLogger('s3tup')

def s3tup(config, access_key_id=None, secret_access_key=None, rsync_only=False,):

    log.info('**** s3tup ****')
    config = yaml.load(file(config))

    if access_key_id is None:
        try: access_key_id = os.environ['AWS_ACCESS_KEY_ID']
        except KeyError: raise Exception('You must supply an aws access key id.')

    if secret_access_key is None:
        try: secret_access_key = os.environ['AWS_SECRET_ACCESS_KEY']
        except KeyError: raise Exception('You must supply an aws secret access key.')

    conn = Connection(access_key_id, secret_access_key)
    bf = BucketFactory(conn)
    for c in config:
        b = bf.make_bucket(**c)
        b.sync(rsync_only)
    
    from connection import stats
    log.info('request totals:')
    for k,v in stats.iteritems():
        if v > 0:
            log.info('- {}: {}'.format(k,v))


KEY_ATTRS = (
    'cache_control',
    'content_disposition',
    'content_encoding',
    'content_type',
    'content_language',
    'expires',
    'reduced_redundancy',
    'acl',
    'canned_acl',
    'encrypt',
    'redirect',
    'metadata',
)

KEY_HEADERS = (
    'cache_control',
    'content_disposition',
    'content_encoding',
    'content_type',
    'content_language',
    'expires',
)

BUCKET_ATTRS = (
    'bucket',
    'canned_acl',
    'cors',
    'key_config',
    'lifecycle',
    'region',
    'rsync',
    'website',
    #'acl',
    #'logging',
    #'notification',
    #'policy',
    #'request_payment',
    #'tagging',
    #'versioning',
)


from base64 import b64encode
from email.utils import formatdate
import logging
import hashlib
import hmac
import urllib

from bs4 import BeautifulSoup
import requests

log = logging.getLogger('s3tup.connection')
stats = {'GET':0, 'POST':0, 'PUT':0, 'DELETE':0, 'HEAD':0}

class Connection(object):

    def __init__(self, access_key_id, secret_access_key):
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key

    def make_request(self, method, bucket, key=None, params=None, headers={}, data=None):

        # Construct target url
        url = 'http://{}.s3.amazonaws.com'.format(bucket)
        url += '/{}'.format(key) if key is not None else '/'
        try: url += '?{}'.format(urllib.urlencode(params))
        except TypeError:
            url += '?{}'.format(params) if params is not None else ''

        # Lowercase all header keys
        headers = {k.lower():v for k,v in headers.iteritems()}

        headers['host'] = '{}.s3.amazonaws.com'.format(bucket)

        if 'content-md5' in headers:
            md5 = headers['content-md5']
        elif data is not None:
            m = hashlib.md5()
            m.update(data)
            md5 = b64encode(m.digest())
            headers['content-md5'] = md5
        else:
            md5 = ''

        try: content_type = headers['content-type']
        except KeyError: content_type = ''

        try: date = headers['x-amz-date']
        except KeyError:
            date = formatdate(timeval=None, localtime=False, usegmt=True)
            headers['x-amz-date'] = date

        # Construct canonicalized resource string
        canonicalized_resource = '/' + bucket
        canonicalized_resource += '/{}'.format(key) if key is not None else '/'
        if isinstance(params, basestring):
            canonicalized_resource += '?{}'.format(params)

        # Construct canonicalized amz headers string
        canonicalized_amz_headers = ''
        amz_keys = sorted([k for k in headers.keys() if k.startswith('x-amz-')])
        for k in amz_keys:
            canonicalized_amz_headers += '{}:{}\n'.format(k, headers[k].strip())

        # Construct string to sign
        string_to_sign = method.upper() + '\n'
        string_to_sign += md5 + '\n'
        string_to_sign += content_type + '\n'
        string_to_sign += '\n' # date is always done through x-amz-date header
        string_to_sign += canonicalized_amz_headers + canonicalized_resource

        # Create signature
        h = hmac.new(self.secret_access_key, string_to_sign, hashlib.sha1)
        signature = b64encode(h.digest())

        # Set authorization header
        headers['Authorization'] = 'AWS {}:{}'.format(self.access_key_id, signature)

        # Logging stuff
        log.debug('{}: {}'.format(method, url))
        log.debug('headers:')
        for k, v in headers.iteritems():
            log.debug(' {}: {}'.format(k, v))

        r = requests.request(method, url, headers=headers, data=data)
        log.debug('response: {}\n'.format(r.status_code))
        stats[method.upper()] += 1
        
        # Handle errors
        if r.status_code/100 != 2:
            log.error("s3 responded with non 2xx response code!")
            log.error("response code: {}".format(r.status_code))
            log.error("request headers:")
            for k,v in r.request.headers.iteritems():
                log.error("  {}: {}".format(k,v))
            try:
                soup = BeautifulSoup(r.text)
                error = soup.find('error')
                for c in error.children:
                    log.error('{}: {}'.format(c.name, c.text))
            except:pass
            raise Exception("s3 responded with non 2xx response code")

        return r




import logging
import mimetypes
import os

import utils
import constants

log = logging.getLogger('s3tup.key')

class KeyFactory(object):

    def __init__(self, conn, bucket_name, config=[]):
        self.conn = conn
        self.bucket_name = bucket_name

        self.configurators = []
        for c in config:
            self.configurators.append(KeyConfigurator(**c))

    def make_key(self, key_name):
        key = Key(self.conn, key_name, self.bucket_name)
        for c in self.configurators:
            if c.effects_key(key.name):
                key = c.configure(key)
        return key

class KeyConfigurator(object):

    def __init__(self, **kwargs):

        # instantiating objects in constructor is an antipattern
        # but whatever
        self.matcher = utils.Matcher(
                patterns=kwargs.pop('patterns', None),
                ignore_patterns=kwargs.pop('ignore_patterns', None),
                regexes=kwargs.pop('regexes', None),
                ignore_regexes=kwargs.pop('ignore_regexes', None),
        )

        for k, v in kwargs.iteritems():
            if k in constants.KEY_ATTRS:
                self.__dict__[k] = v

    def effects_key(self, key_name):
        return self.matcher.match(key_name)

    def configure(self, key):
        for attr in constants.KEY_ATTRS:
            if attr in self.__dict__ and attr != 'metadata':
                key.__dict__[attr] = self.__dict__[attr]

        try: key.metadata.update(self.metadata)
        except AttributeError: pass

        return key

class Key(object):
    
    def __init__(self, conn, name, bucket_name, **kwargs):
        self.conn = conn
        self.name = name
        self.bucket = bucket_name

        # set defaults for required attributes
        self.reduced_redundancy = kwargs.pop('reduced_redundancy', False)
        self.encrypt = kwargs.pop('encrypt', False)
        self.metadata = kwargs.pop('metadata', {})

        for attr in kwargs: # for each parameter passed in to this constuctor
            if attr in constants.KEY_ATTRS: # check if the parameter is allowed
                self.__dict__[attr] = kwargs[attr]
            else:
                raise TypeError("Key.__init__() got an unexpected keyword argument\
                                 '{}'".format(attr))

    @property
    def headers(self):
        headers = {}

        try: headers['x-amz-website-redirect-location'] = self.redirect
        except AttributeError: pass

        try: headers['x-amz-acl'] = self.canned_acl
        except AttributeError: pass

        try:
            if self.reduced_redundancy:
                headers['x-amz-storage-class'] = 'REDUCED_REDUNDANCY'
        except AttributeError: pass

        try: 
            if self.encrypt:
                headers['x-amz-server-side-encryption'] = 'AES256'
        except AttributeError: pass

        for k, v in self.metadata.iteritems():
            headers['x-amz-meta-' + k] = v

        for k in constants.KEY_HEADERS:
            try:
                if self.__dict__[k] is not None:
                    headers[k.replace('_', '-')] = self.__dict__[k]
            except KeyError: pass

        # Guess content-type
        if 'content-type' not in headers:
            content_type_guess = mimetypes.guess_type(self.name)[0]
            if content_type_guess is not None:
                headers['content-type'] = content_type_guess

        return headers

    def sync(self):
        log.info("syncing key '{}'...".format(self.name))

        headers = self.headers
        headers['x-amz-copy-source'] = '/' + self.bucket + '/' + self.name
        headers['x-amz-metadata-directive'] = 'REPLACE'

        self.conn.make_request('PUT', self.bucket, self.name, headers=headers)
        self.sync_acl()

    def rsync(self, filename):
        log.info("uploading key '{}' from '{}'...".format(self.name, filename))

        headers = self.headers
        headers['content-md5'] = utils.file_md5(filename)
        headers['content-length'] = os.path.getsize(filename)
        data = open(filename, 'rb').read()
        self.conn.make_request('PUT', self.bucket, self.name, headers=headers, 
                                data=data)
        self.sync_acl()

    def sync_acl(self):
        try:
            self.conn.make_request('PUT', self.bucket, self.name, 'acl',
                                   data=self.acl)
        except AttributeError: pass




from fnmatch import fnmatch
from base64 import b64encode
import re
import hashlib
import binascii

from bs4 import BeautifulSoup

class Matcher(object):
    
    def __init__(self, patterns=None, ignore_patterns=None, regexes=None, ignore_regexes=None):
        self.patterns = patterns
        self.ignore_patterns = ignore_patterns
        self.regexes = regexes
        self.ignore_regexes = ignore_regexes

    def match(self, s):
        matched = True if self.patterns is None and self.regexes is None else False

        if self.patterns is not None:
            for pattern in self.patterns:
                if fnmatch(s, pattern):
                    matched = True
                    break

        if self.regexes is not None and matched is not True:
            for regex in self.regexes:
                if re.search(regex, s):
                    matched = True
                    break

        if self.ignore_patterns is not None and matched is True:
            for pattern in self.ignore_patterns:
                if fnmatch(s, pattern):
                    return False

        if self.ignore_regexes is not None and matched is True:
            for regex in self.ignore_regexes:
                if re.search(regex, s):
                    return False

        return matched


def file_md5(filename):
    m = hashlib.md5()
    with open(filename,'rb') as f:
        while True:
            buf=f.read(8192)
            if not buf: break
            m.update(buf)
    return b64encode(m.digest()).strip()

def list_bucket(conn, bucket_name, marker=None):
    marker = None
    more_results = True
    while more_results:
        params = {'marker': marker, 'max-keys': 10}
        r = conn.make_request('GET', bucket_name, params=params)
        soup = BeautifulSoup(r.text)
        root = soup.find('listbucketresult')
        
        for c in root.find_all('contents'):
            key = c.find('key').text
            etag_hex = c.find('etag').text.replace('"', '')
            etag_bin = binascii.unhexlify(etag_hex)
            etag = binascii.b2a_base64(etag_bin).strip()
            yield {'name': key, 'etag': etag}
            marker = key

        more_results = root.find('istruncated').text == 'true'
        


import argparse
import logging
from s3tup import s3tup

parser = argparse.ArgumentParser(description='s3tup')
parser.add_argument(
	'config',
	help='path to configuration file'
	)
parser.add_argument(
	'--access_key_id',
	help='your aws access key id'
)
parser.add_argument(
	'--secret_access_key',
	help='your aws secret access key'
)
parser.add_argument(
	'--rsync_only',
	action='store_true',
	help='only run rsync, do not sync other keys'
)
verbosity_group = parser.add_mutually_exclusive_group()
verbosity_group.add_argument(
	"-v",
	"--verbose",
	action="store_true",
    help="run in verbose mode"
)
verbosity_group.add_argument(
	"--debug",
	action="store_true",
    help="run in debug mode"
)

def main():
	args = parser.parse_args()

	if args.debug:
		logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
	elif args.verbose:
		logging.basicConfig(format="%(message)s", level=logging.INFO)

	s3tup(args.config, args.access_key_id, args.secret_access_key, args.rsync_only)


import logging

import os
from utils import Matcher, list_bucket, file_md5

log = logging.getLogger('s3tup')

def rsync(key_factory, src='', delete=False, patterns=None,
          ignore_patterns=None, regexes=None, ignore_regexes=None):

    matcher = Matcher(patterns, ignore_patterns, regexes, ignore_regexes)
    conn = key_factory.conn
    bucket = key_factory.bucket_name
    log.info("rsyncing folder '{}' with bucket '{}'...".format(src, bucket))

    class RsyncInternalKey(object):
        def __init__(self, name):
            self.name = name
            self.local = False
            self.remote = False
        @property
        def local_path(self):
            return os.path.join(src, self.name)
        @property
        def local_md5(self):
            return file_md5(self.local_path)
        @property
        def new(self):
            return self.local == True and self.remote == False
        @property
        def removed(self):
            return self.local == False and self.remote == True
        @property
        def modified(self):
            if self.local == False or self.remote == False:
                return False
            return self.local_md5 != self.remote_md5
        @property
        def unmodified(self):
            if self.local == False or self.remote == False:
                return False
            return self.local_md5 == self.remote_md5
        def rsync(self):
            key = key_factory.make_key(self.name)
            key.rsync(self.local_path)
        def __lt__(self, other):
            return self.name < other.name


    all_keys = {}

    # Add all local keys
    for path, dirs, files in os.walk(src):
        for f in files:
            full_path = os.path.join(path, f)
            rel_path = os.path.relpath(full_path, src)
            if matcher.match(rel_path):
                k = RsyncInternalKey(rel_path)
                k.local = True
                all_keys[k.name] = k

    # Add all remote keys
    for list_key in list_bucket(key_factory.conn, key_factory.bucket_name):
        try: k = all_keys[list_key['name']]
        except KeyError: k = RsyncInternalKey(list_key['name'])
        k.remote = True
        k.remote_md5 = list_key['etag']
        all_keys[k.name] = k

    new_keys = [k for k,v in all_keys.iteritems() if v.new]
    removed_keys = [k for k,v in all_keys.iteritems() if v.removed]
    modified_keys = [k for k,v in all_keys.iteritems() if v.modified]
    unmodified_keys =[k for k,v in all_keys.iteritems() if v.unmodified]

    if delete == True:
        for i in xrange(0, len(removed_keys), 1000):
            data = """<?xml version="1.0" encoding="UTF-8"?>
                      <Delete>
                      <Quiet>true</Quiet>\n"""
            for k in removed_keys[i:i+1000]:
                log.info('removed: {}'.format(k))
                data += '<Object><Key>{}</Key></Object>\n'.format(k)
            data += '</Delete>'
            conn.make_request('POST', bucket, None, 'delete', data=data)
    else:
        unmodified_keys.extend(removed_keys)
        removed_keys = []

    for k in new_keys:
        log.info('new: {}, uploading now...'.format(k)) 
        key = all_keys[k]
        key.rsync()

    for k in modified_keys:
        log.info('modified: {}, uploading now...'.format(k)) 
        key = all_keys[k]
        key.rsync()

    log.info('rsync complete!')
    log.info('{} new, {} removed, {} modified, {} unmodified'.format(
             len(new_keys), len(removed_keys), len(modified_keys),
             len(unmodified_keys)))

    return {'new': new_keys, 
            'removed': removed_keys, 
            'modified': modified_keys, 
            'unmodified': unmodified_keys}


from setuptools import setup

setup(name='s3tup',
      version='0.0.1',
      description='Declarative configuration management tool for amazon s3',
      url='http://github.com/HeyImAlex/s3tup',
      author='Alex Guerra',
      author_email='alex@heyimalex.com',
      license='MIT',
      packages=['s3tup'],
      install_requires=[
          'requests',
          'argparse',
          'pyyaml',
          'beautifulsoup4',
      ],
      entry_points = {
        'console_scripts': ['s3tup=s3tup.cli:main'],
      },
      zip_safe=False)


import logging

from connection import Connection
from key import KeyFactory
from rsync import rsync
import utils
import constants

log = logging.getLogger('s3tup.bucket')

class BucketFactory(object):

    def __init__(self, conn=None):
        self.conn = conn

    def make_bucket(self, **kwargs):
        bucket_name = kwargs.pop('bucket')

        try:
            access_key_id = kwargs.pop('access_key_id')
            secret_access_key = kwargs.pop('secret_access_key')
            conn = Connection(access_key_id, secret_access_key)
        except KeyError:
            conn = self.conn

        bucket = Bucket(conn, bucket_name, **kwargs)
        return bucket
        

class Bucket(object):

    def __init__(self, conn, name, **kwargs):
        self.conn = conn
        self.name = name

        for attr in kwargs:
            if attr in constants.BUCKET_ATTRS:
                self.__dict__[attr] = kwargs[attr]
            else:
                raise TypeError("Bucket.__init__() got an unexpected keyword\
                                 argument '{}'".format(attr))

    def sync(self, rsync_only=False):

        log.info("syncing bucket '{}'...".format(self.name))

        # create bucket
        headers = {}
        try: headers['x-amz-acl'] = self.canned_acl
        except AttributeError: pass
        try:
            data = """<CreateBucketConfiguration 
                       xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
                      <LocationConstraint>{}</LocationConstraint> 
                      </CreateBucketConfiguration >""".format(self.region)
        except AttributeError:
            data = None
        self.conn.make_request('PUT', self.name, headers=headers, data=data)

        # TODO- implement all of these methods
        # self.sync_acl()
        # self.sync_logging()
        # self.sync_notification()
        # self.sync_policy()
        # self.sync_tagging()
        # self.sync_versioning()
        self.sync_lifecycle()
        self.sync_cors()
        self.sync_website()

        try: # create key factory from key_config if it's set
            kf = KeyFactory(self.conn, self.name, self.key_config)
        except AttributeError: kf = KeyFactory(self.conn, self.name)

        if 'rsync' in self.__dict__:
            rs_out = rsync(kf, **self.rsync)
            unmodified = rs_out['unmodified']
        else:
            unmodified = [k['name'] for k in utils.list_bucket(self.conn, self.name)]

        if not rsync_only:
            if 'key_config' in self.__dict__:
                for k in unmodified:
                    key = kf.make_key(k)
                    key.sync()

        log.info("bucket '{}' sucessfully synced!\n".format(self.name))

    def sync_cors(self):
        try:
            if self.cors is not None:
                log.info("setting cors configuration...")
                self.conn.make_request('PUT', self.name, None, 'cors',
                                       data=self.cors)
            else:
                log.info("deleting cors configuration...")
                self.conn.make_request('DELETE', self.name, None, 'cors')
        except AttributeError: pass

    def sync_website(self):
        try:
            if self.website is not None:
                log.info("setting website configuration...")
                self.conn.make_request('PUT', self.name, None, 'website',
                                       data=self.website)
            else:
                log.info("deleting website configuration...")
                self.conn.make_request('DELETE', self.name, None, 'website')
        except AttributeError: pass

    def sync_lifecycle(self):
        try:
            if self.lifecycle is not None:
                log.info("setting lifecycle configuration...")
                self.conn.make_request('PUT', self.name, None, 'lifecycle',
                                       data=self.lifecycle)
            else:
                log.info("deleting lifecycle configuration...")
                self.conn.make_request('DELETE', self.name, None, 'lifecycle')
        except AttributeError: pass